/*
 *  802.11 40/104 bit WEP decryptor
 *
 *  Copyright (C) 2004  Christophe Devine
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>

#include "pcap.h"

#define SWAP(x,y) { unsigned char tmp = x; x = y; y = tmp; }

char usage[] =

"\n"
"  airunwep 1.0 - (C) 2004 Christophe Devine\n"
"\n"
"  usage: %s <key> <pcap infile> <pcap outfile>\n"
"\n";

unsigned char llc_header[6] =
{
    0xAA, 0xAA, 0x03, 0x00, 0x00, 0x00
};

unsigned char buffer[65536];

int main( int argc, char *argv[] )
{
    unsigned char K[16];
    unsigned char S[256];
    unsigned char *h80211;
    unsigned char wepkey[13];

    struct pcap_file_header pfh_in;
    struct pcap_file_header pfh_out;
    struct pcap_pkthdr pkh;

    int weplen, infile, outfile;
    int n, i, j, d, p1, p2;
    long int cnt1, cnt2;
    time_t tm_prev;
    char *s;

    /* check the arguments */

    if( argc != 4 )
    {
    usage:
        printf( usage, argv[0] );
        return( 1 );
    }

    s = argv[1];
    weplen = 0;

    while( sscanf( s, "%x", &n ) == 1 )
    {
        if( n < 0 || n > 255 )
            goto usage;

        wepkey[weplen++] = n;

        if( weplen > 13 ) break;

        if( ! ( s = strchr( s, ':' ) ) )
            break;

        s++;
    }

    if( weplen != 5 && weplen != 13 )
        goto usage;

    /* open the input file and check the pcap header */

    infile = open( argv[2], O_RDONLY | O_LARGEFILE );

    if( infile < 0 )
    {
        perror( "open(pcap infile)" );
        return( 1 );
    }

    n = sizeof( struct pcap_file_header );

    if( read( infile, &pfh_in, n ) != n )
    {
        perror( "read(infile pcap header)" );
        return( 1 );
    }

    if( pfh_in.magic != TCPDUMP_MAGIC )
    {
        fprintf( stderr, "wrong magic from pcap file header\n"
                         "(got 0x%08X, expected 0x%08X)\n",
                         pfh_in.magic, TCPDUMP_MAGIC );
        return( 1 );
    }

    if( pfh_in.linktype != LINKTYPE_IEEE802_11 &&
        pfh_in.linktype != LINKTYPE_PRISM_HEADER )
    {
        fprintf( stderr, "unsupported hardware link type %d\n",
                 pfh_in.linktype );
        return( 1 );
    }

    /* attempt to open or create the output file */

    n = sizeof( struct pcap_file_header );

    if( ( outfile = open( argv[3], O_LARGEFILE | O_RDWR ) ) < 0 )
    {
        if( ( outfile = open( argv[3], O_LARGEFILE | O_RDWR |
                                       O_CREAT, 0644 ) ) < 0 )
        {
            perror( "creat(pcap outfile)" );
            return( 1 );
        }

        /* newly created file, so append the pcap header */

        pfh_out.magic           = TCPDUMP_MAGIC;
        pfh_out.version_major   = PCAP_VERSION_MAJOR;
        pfh_out.version_minor   = PCAP_VERSION_MINOR;
        pfh_out.thiszone        = 0;
        pfh_out.sigfigs         = 0;
        pfh_out.snaplen         = 65535;
        pfh_out.linktype        = LINKTYPE_ETHERNET;

        write( outfile, &pfh_out, n );
    }
    else
    {
        /* existing file, check the pcap header contents */

        if( read( outfile, &pfh_out, n ) != n )
        {
            perror( "read(outfile pcap header)" );
            return( 1 );
        }

        if( pfh_out.magic != TCPDUMP_MAGIC )
        {
            fprintf( stderr, "wrong magic from pcap file header\n"
                             "(got 0x%08X, expected 0x%08X)\n",
                             pfh_out.magic, TCPDUMP_MAGIC );
            return( 1 );
        }

        if( pfh_out.linktype != LINKTYPE_ETHERNET )
        {
            fprintf( stderr, "wrong linktype from pcap file header\n"
                             "(got %d, expected LINKTYPE_ETHERNET)\n",
                     pfh_out.linktype );
            return( 1 );
        }

        if( lseek64( outfile, 0, SEEK_END ) < 0 )
        {
            perror( "lseek64(SEEK_END)" );
            return( 1 );
        }
    }

    memcpy( K + 3, wepkey, weplen );

    tm_prev = time( NULL );

    cnt1 = cnt2 = 0;

    while( 1 )
    {
        /* every 1s, update the display */

        if( time( NULL ) - tm_prev >= 1 )
        {
            tm_prev = time( NULL );
            printf( "\r\33[2KReading packets: total = %ld"
                    ", decrypted = %ld\r", cnt1, cnt2 );
            fflush( stdout );
        }

        /* read one packet */

        n = sizeof( pkh );

        if( read( infile, &pkh, n ) != n )
            break;

        if( read( infile, buffer, pkh.len ) != (int) pkh.len )
            break;

        cnt1++;

        h80211 = buffer;

        if( pfh_in.linktype == LINKTYPE_PRISM_HEADER )
        {
            /* remove the prism header if necessary */

            n = *(int *)( h80211 + 4 );
            h80211 += n; pkh.len -= n;

            if( pkh.len <= 0 )
            {
                fprintf( stderr, "\nFIXME: wrong prism header size ?\n" );
                continue;
            }
        }

        if( pkh.len < 40 ) continue;

        /* only consider encrypted data packets */

        if( ( h80211[0] & 0x0C ) != 0x08 ) continue;
        if( ( h80211[1] & 0x40 ) != 0x40 ) continue;

        /* attempt to decrypt the packet */

        memcpy( K, h80211 + 24, 3 );

        for( i = 0; i < 256; i++ )
            S[i] = i;

        p1 = p2 = j = d = 0;

        for( i = 0; i < 256; i++ )
        {
            j = ( j + S[i] + K[d]) & 0xFF;
            SWAP( S[i], S[j] );
            if( ++d >= ( 3 + weplen ) ) d = 0;
        }

        for( i = 1, j = 0; i < 7; i++ )
        {
            p1 = ( p1 + 1     ) & 0xFF;
            p2 = ( p2 + S[p1] ) & 0xFF;
            SWAP( S[p1], S[p2] );
            h80211[28 + i - 1] ^= S[(S[p1] + S[p2]) & 0xFF];

            if( h80211[28 + i - 1] != llc_header[i - 1] )
                break;
        }

        if( i < 7 ) continue;

        /* wepkey ok, so decrypt the rest of the packet */

        for( ; i <= (int) pkh.len - 32; i++ )
        {
            p1 = ( p1 + 1     ) & 0xFF;
            p2 = ( p2 + S[p1] ) & 0xFF;
            SWAP( S[p1], S[p2] );
            h80211[28 + i - 1] ^= S[(S[p1] + S[p2]) & 0xFF];
        }

        /* remove the 802.11/WEP/LLC header & padding/FCS tail */

        pkh.len    -= 28 + 6 + ( h80211[27] & 0x3F ) + 4;
        pkh.caplen -= 28 + 6 + ( h80211[27] & 0x3F ) + 4;

        /* also create the Ethernet link layer (MAC dst+src) */

        pkh.len    += 12;
        pkh.caplen += 12;

        memcpy( h80211 + 34 - 12, h80211 + 4, 12 );

        if( write( outfile, &pkh,  sizeof( pkh ) ) != sizeof( pkh ) ||
            write( outfile, h80211 + 22, pkh.len ) != (int) pkh.len )
        {
            perror( "\nwrite" );
            return( 1 );
        }

        cnt2++;
    }

    printf( "\r\33[2KReading packets: total = %ld"
            ", decrypted = %ld\n", cnt1, cnt2 );

    return( 0 );
}
